@@ -3,7 +3,7 @@
 #
 # ibus-skk - The SKK engine for IBus
 #
-# Copyright (C) 2009 Daiki Ueno <ueno@unixuser.org>
+# Copyright (C) 2009-2010 Daiki Ueno <ueno@unixuser.org>
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -24,6 +24,7 @@ from __future__ import with_statement
 import itertools
 import os.path
 import socket
+import re
 
 # Converted from skk-rom-kana-base-rule in skk-vars.el.
 ROM_KANA_RULE = {
@@ -300,6 +301,11 @@ KUTOUTEN_EN_JP = range(4)
 
 KUTOUTEN_RULE = ((u'。', u'、'), (u'．', u'，'), (u'。', u'，'), (u'．', u'、'))
 
+AUTO_START_HENKAN_KEYWORDS = (u'を', u'、', u'。', u'．', u'，', u'？', u'」',
+                              u'！', u'；', u'：', u')', u';', u':', u'）',
+                              u'”', u'】', u'』', u'》', u'〉', u'｝', u'］',
+                              u'〕', u'}', u']', u'?', u'.', u',', u'!')
+
 CONV_STATE_NONE, \
 CONV_STATE_START, \
 CONV_STATE_SELECT = range(3)
@@ -332,6 +338,7 @@ class DictBase(object):
     ENCODING = 'EUC-JP'
 
     def split_candidates(self, line):
+        '''Parse a single candidate line into a list of candidates.'''
         def seperate_annotation(candidate):
             index = candidate.find(u';')
             if index >= 0:
@@ -341,6 +348,7 @@ class DictBase(object):
         return map(seperate_annotation, line.strip()[1:-1].split(u'/'))
 
     def join_candidates(self, candidates):
+        '''Make a single candidate line from a list of candidates.'''
         def append_annotation(candidate_annotation):
             candidate, annotation = candidate_annotation
             if annotation is not None:
@@ -350,16 +358,26 @@ class DictBase(object):
         return u'/'.join(map(append_annotation, candidates))
 
     def reload(self):
+        '''Reload the dictionary.'''
         raise NotImplemented
 
     def lookup(self, midasi, okuri=False):
+        '''Lookup MIDASI in the dictionary.'''
         raise NotImplemented
         
     def completer(self, midasi):
+        '''Return an iterator to complete MIDASI.'''
         raise NotImplemented
 
-    def save(self):
-        raise NotImplemented
+class EmptyDict(DictBase):
+    def reload(self):
+        pass
+
+    def lookup(self, midasi, okuri=False):
+        return list()
+        
+    def completer(self, midasi):
+        return iter(list())
 
 class SysDict(DictBase):
     def __init__(self, path, encoding=DictBase.ENCODING):
@@ -483,6 +501,18 @@ class SysDict(DictBase):
 
 class UsrDict(DictBase):
     PATH = '~/.skk-ibus-jisyo'
+    HISTSIZE = 128
+
+    __encoding_to_coding_system = {
+        'UTF-8': 'utf-8',
+        'EUC-JP': 'euc-jp',
+        'Shift_JIS': 'shift_jis',
+        'ISO-2022-JP': 'iso-2022-jp'
+        }
+
+    __coding_system_to_encoding = dict()
+    for encoding, coding_system in __encoding_to_coding_system.items():
+        __coding_system_to_encoding[coding_system] = encoding
 
     def __init__(self, path=PATH, encoding=DictBase.ENCODING):
         self.__path = os.path.expanduser(path)
@@ -491,30 +521,73 @@ class UsrDict(DictBase):
 
     path = property(lambda self: self.__path)
 
+    __coding_cookie_pattern = \
+        re.compile('\A\s*;+\s*-\*-\s*coding:\s*(\S+?)\s*-\*-')
     def reload(self):
         self.__dict = dict()
-        with open(self.__path, 'a+') as fp:
-            for line in fp:
-                if line.startswith(';'):
-                    continue
-                line = line.decode(self.__encoding)
-                midasi, candidates = line.split(' ', 1)
-                self.__dict[midasi] = self.split_candidates(candidates)
+        try:
+            with open(self.__path, 'a+') as fp:
+                line = fp.readline()
+                if line:
+                    match = re.match(self.__coding_cookie_pattern, line)
+                    if match:
+                        encoding = self.__coding_system_to_encoding.\
+                            get(match.group(1))
+                        if encoding:
+                            self.__encoding = encoding
+                fp.seek(0)
+                for line in fp:
+                    if line.startswith(';'):
+                        continue
+                    line = line.decode(self.__encoding)
+                    midasi, candidates = line.split(' ', 1)
+                    self.__dict[midasi] = self.split_candidates(candidates)
+            self.__read_only = False
+        except Exception:
+            self.__read_only = True
         self.__dict_changed = False
+        self.__selection_history = list()
 
+    read_only = property(lambda self: self.__read_only)
+
+    def lookup(self, midasi, okuri=False):
+        return self.__dict.get(midasi, list())
+
+    def completer(self, midasi):
+        for _midasi in self.__selection_history:
+            if _midasi.startswith(midasi):
+                yield _midasi
+        # TODO: complete from __dict?
+        
     def save(self):
-        if not self.__dict_changed:
+        '''Save the changes to the user dictionary.'''
+        if not self.__dict_changed or self.__read_only:
             return
         with open(self.__path, 'w+') as fp:
+            coding_system = self.__encoding_to_coding_system.\
+                get(self.__encoding)
+            if coding_system:
+                fp.write(';;; -*- coding: %s -*-\n' % coding_system)
             for midasi in sorted(self.__dict):
                 line = midasi + u' /' + \
                     self.join_candidates(self.__dict[midasi]) + '/\n'
                 fp.write(line.encode(self.__encoding))
 
-    def lookup(self, midasi, okuri=False):
-        return self.__dict.get(midasi, list())
-        
     def select_candidate(self, midasi, candidate):
+        '''Mark CANDIDATE was selected as the conversion result of MIDASI.'''
+        del(self.__selection_history[self.HISTSIZE:])
+        _midasi = None
+        for index, _midasi in enumerate(self.__selection_history):
+            if _midasi == midasi:
+                if index > 0:
+                    first = self.__selection_history[0]
+                    self.__selection_history[0] =\
+                        self.__selection_history[index]
+                    self.__selection_history[index] = first
+                break
+        if _midasi is not midasi:
+            self.__selection_history.insert(0, midasi)
+
         if midasi not in self.__dict:
             self.__dict[midasi] = list()
         elements = self.__dict[midasi]
@@ -592,41 +665,78 @@ def compile_rom_kana_rule(rule):
         _compile_rom_kana_rule(tree, input_state, rule[input_state])
     return tree
 
-class CandidateSelectorBase(object):
-    def __init__(self):
+class CandidateSelector(object):
+    PAGE_SIZE = 7
+    PAGINATION_START = 4
+
+    def __init__(self, page_size=PAGE_SIZE, pagination_start=PAGINATION_START):
+        self.__page_size = page_size
+        self.__pagination_start = pagination_start
         self.set_candidates(list())
 
+    page_size = property(lambda self: self.__page_size)
+    pagination_start = property(lambda self: self.__pagination_start)
+
     def set_candidates(self, candidates):
+        '''Set the list of candidates.'''
         self.__candidates = candidates
-        self.__candidate_index = -1
+        self.__index = -1
 
-    def next_candidate(self):
-        if self.__candidates:
-            self.__candidate_index += 1
-            self.__candidate_index %= len(self.__candidates)
-            return self.__candidates[self.__candidate_index]
+    def next_candidate(self, move_over_pages=True):
+        '''Move the cursor forward.  If MOVE_OVER_PAGES is True, skip
+        to the next page instead of the next candidate.'''
+        if move_over_pages and self.__index >= self.__pagination_start:
+            index = self.__index + self.__page_size
+            # Place the cursor at the beginning of the page.
+            index -= (index - self.__pagination_start) % self.__page_size
+        else:
+            index = self.__index + 1
+        self.set_index(index)
+        return self.candidate()
 
-    def previous_candidate(self):
-        if self.__candidates:
-            self.__candidate_index -= 1
-            self.__candidate_index %= len(self.__candidates)
-            return self.__candidates[self.__candidate_index]
+    def previous_candidate(self, move_over_pages=True):
+        '''Move the cursor forward.  If MOVE_OVER_PAGES is
+        True, skip to the previous page instead of the previous candidate.'''
+        if move_over_pages and self.__index >= self.__pagination_start:
+            index = self.__index - self.__page_size
+            # Place the cursor at the beginning of the page.
+            index -= (index - self.__pagination_start) % self.__page_size
+        else:
+            index = self.__index - 1
+        self.set_index(index)
+        return self.candidate()
+
+    def candidate(self):
+        '''Return the current candidate.'''
+        if self.__index < 0:
+            return None
+        return self.__candidates[self.__index] + (True,)
+
+    def index(self):
+        '''Return the current candidate index.'''
+        return self.__index
+
+    def set_index(self, index):
+        '''Set the current candidate index.'''
+        if 0 <= index and index < len(self.__candidates):
+            self.__index = index
+        else:
+            self.__index = -1
 
 class Context:
-    def __init__(self, usrdict, sysdict):
+    def __init__(self, usrdict, sysdict, candidate_selector):
         '''Create an SKK context.
 
         USRDICT is a user dictionary and SYSDICT is a system
         dictionary (or a connection to SKKServ).'''
         self.__usrdict = usrdict
         self.__sysdict = sysdict
+        self.__candidate_selector = candidate_selector
 
         self.__rom_kana_rule_tree = compile_rom_kana_rule(ROM_KANA_RULE)
-        self.set_candidate_selector(CandidateSelectorBase())
         self.reset()
-        self.activate_input_mode(INPUT_MODE_HIRAGANA)
         self.kutouten_type = KUTOUTEN_JP
-        self.__comphist = dict()
+        self.auto_start_henkan_keywords = AUTO_START_HENKAN_KEYWORDS
 
     def __check_dict(self, _dict):
         if not isinstance(_dict, DictBase):
@@ -646,36 +756,38 @@ class Context:
     sysdict = property(lambda self: self.__sysdict, set_sysdict)
 
     def reset(self):
-        '''Reset the internal state of the context.'''
+        '''Reset the internal state of the context.  The initial state
+        is INPUT_MODE_HIRAGANA/CONV_STATE_NONE.'''
+        # Current midasi in conversion.
+        self.__midasi = None
+
+        self.__candidate_selector.set_candidates(list())
+
+        # Whether or not we are in the abbrev mode.
+        self.__abbrev = False
+
+        self.__conv_state = CONV_STATE_NONE
+
+        self.__completer = None
+
+        # Last used keyword which triggered auto-start-henkan.
+        self.__auto_start_henkan_keyword = None
+
         # rom-kana state is either None or a tuple
         #
         # (OUTPUT, PENDING, TREE)
         #
         # where OUTPUT is a kana string, PENDING is a string in
         # rom-kana conversion, and TREE is a subtree of
-        # __ROM_KANA_RULE_TREE.
+        # __rom_kana_rule_tree.
         #
-        # See __convert_rom_kana for the state transition algorithm.
-        self.__rom_kana_state = None
+        # See __convert_rom_kana() for the state transition algorithm.
         self.__okuri_rom_kana_state = None
-
-        # kana-kan state is either None or a tuple
-        #
-        # (MIDASI, CANDIDATE)
-        #
-        # where MIDASI is a kana string used as a key for dict search
-        # and CANDIDATE is the current candidate selected (if any).
-        self.__kana_kan_state = None
-
-        self.__abbrev = False
-
-        self.__conv_state = CONV_STATE_NONE
-        self.clear_candidates()
-
-        self.__comp_state = None
+        self.activate_input_mode(INPUT_MODE_HIRAGANA)
 
     conv_state = property(lambda self: self.__conv_state)
     input_mode = property(lambda self: self.__input_mode)
+    abbrev = property(lambda self: self.__abbrev)
 
     def __hiragana_to_katakana(self, kana):
         diff = ord(u'ア') - ord(u'あ')
@@ -698,30 +810,39 @@ class Context:
         self.__input_mode = input_mode
         if self.__input_mode in (INPUT_MODE_HIRAGANA, INPUT_MODE_KATAKANA):
             self.__rom_kana_state = (u'', u'', self.__rom_kana_rule_tree)
+        else:
+            self.__rom_kana_state = None
 
     def kakutei(self):
         '''Fix the current candidate as a commitable string.'''
-        input_mode = self.__input_mode
-        if self.__kana_kan_state:
-            candidate = self.__kana_kan_state[1]
+        if self.__midasi:
+            candidate = self.__candidate_selector.candidate()
             if candidate:
                 output = candidate[0]
                 if self.__okuri_rom_kana_state:
                     output += self.__okuri_rom_kana_state[0]
-                self.__usrdict.select_candidate(self.__kana_kan_state[0],
-                                                candidate)
+                if candidate[2]:
+                    self.__usrdict.select_candidate(self.__midasi,
+                                                    candidate[:2])
             else:
                 output = self.__rom_kana_state[0]
         else:
             output = self.__rom_kana_state[0]
+        if self.__auto_start_henkan_keyword:
+            output += self.__auto_start_henkan_keyword
+        input_mode = self.__input_mode
         self.reset()
         self.activate_input_mode(input_mode)
         return output
 
     def press_key(self, key):
-        '''Process a key press event KEY and return a committable
-        string (if any).  KEY is in the format of
-        ["ctrl+"]["shift+"]<lower case ASCII letter>.'''
+        '''Process a key press event KEY.
+
+        KEY is in the format of ["ctrl+"]["shift+"]<lower case ASCII letter>.
+
+        The return value is a tuple (HANDLED, OUTPUT) where HANDLED is
+        True if the event was handled internally (otherwise False),
+        and OUTPUT is a committable string (if any).'''
         keyval = key
         is_ctrl = keyval.startswith('ctrl+')
         if is_ctrl:
@@ -736,8 +857,9 @@ class Context:
 
         if key == 'ctrl+g':
             if self.__conv_state in (CONV_STATE_NONE, CONV_STATE_START):
+                input_mode = self.__input_mode
                 self.reset()
-                self.activate_input_mode(self.__input_mode)
+                self.activate_input_mode(input_mode)
             else:
                 if self.__okuri_rom_kana_state:
                     self.__rom_kana_state = (self.__rom_kana_state[0] + \
@@ -745,10 +867,10 @@ class Context:
                                              u'', self.__rom_kana_rule_tree)
                     self.__okuri_rom_kana_state = None
                 # Stop kana-kan conversion.
-                self.__kana_kan_state = None
-                self.clear_candidates()
+                self.__midasi = None
+                self.__candidate_selector.set_candidates(list())
                 self.__conv_state = CONV_STATE_START
-            return u''
+            return (True, u'')
 
         if self.__conv_state == CONV_STATE_NONE:
             input_mode = INPUT_MODE_TRANSITION_RULE.get(key, dict()).\
@@ -756,18 +878,22 @@ class Context:
             if input_mode is not None:
                 self.reset()
                 self.activate_input_mode(input_mode)
-                return u''
+                return (True, u'')
+
+            # Ignore ctrl+key and non-ASCII characters.
+            if is_ctrl or 0x20 > ord(letter) or ord(letter) > 0x7E:
+                return (False, u'')
 
             if self.__input_mode == INPUT_MODE_LATIN:
-                return letter
+                return (True, letter)
             elif self.__input_mode == INPUT_MODE_WIDE_LATIN:
-                return WIDE_LATIN_TABLE[ord(letter)]
+                return (True, WIDE_LATIN_TABLE[ord(letter)])
 
             # Start rom-kan mode.
             if keyval == '/':
                 self.__conv_state = CONV_STATE_START
                 self.__abbrev = True
-                return u''
+                return (True, u'')
 
             if is_shift and keyval.isalpha():
                 self.__conv_state = CONV_STATE_START
@@ -776,8 +902,8 @@ class Context:
                 self.__convert_rom_kana(keyval, self.__rom_kana_state)
             if self.__conv_state == CONV_STATE_NONE and \
                     len(self.__rom_kana_state[1]) == 0:
-                return self.kakutei()
-            return u''
+                return (True, self.kakutei())
+            return (True, u'')
 
         elif self.__conv_state == CONV_STATE_START:
             input_mode = INPUT_MODE_TRANSITION_RULE.get(key, dict()).\
@@ -787,61 +913,61 @@ class Context:
                 kana = self.__hiragana_to_katakana(self.__rom_kana_state[0])
                 self.kakutei()
                 self.activate_input_mode(input_mode)
-                return kana
+                return (True, kana)
             elif self.__input_mode == INPUT_MODE_KATAKANA and \
                     input_mode == INPUT_MODE_HIRAGANA:
                 kana = self.__katakana_to_hiragana(self.__rom_kana_state[0])
                 self.kakutei()
                 self.activate_input_mode(input_mode)
-                return kana
+                return (True, kana)
             elif input_mode is not None and not self.__abbrev:
                 output = self.kakutei()
                 self.activate_input_mode(input_mode)
-                return output
+                return (True, output)
 
             # Start TAB completion.
             if keyval == u'\t':
                 self.__rom_kana_state = self.__convert_nn(self.__rom_kana_state)
-                if self.__comp_state is None:
-                    self.__comp_state = self.__init_completion(\
-                        self.__rom_kana_state[0])
-                try:
-                    self.__rom_kana_state = (self.__comp_state[1].next(),
-                                             u'', u'')
-                except StopIteration:
-                    pass
-                return u''
+                if self.__completer is None:
+                    compkey = self.__rom_kana_state[0]
+                    self.__completer = self.__init_completer(compkey)
+                    self.__completion_seen = set((compkey,))
+                for midasi in self.__completer:
+                    if midasi not in self.__completion_seen:
+                        self.__rom_kana_state = (midasi, u'', u'')
+                        self.__completion_seen.add(midasi)
+                        break
+                return (True, u'')
             # Stop TAB completion.
-            if self.__comp_state is not None and \
-                    len(self.__rom_kana_state[0]) > 0:
-                comphist = self.__comphist.setdefault(self.__comp_state[0],
-                                                      list())
-                try:
-                    comphist.remove(self.__rom_kana_state[0])
-                except:
-                    pass
-                comphist.insert(0, self.__rom_kana_state[0])
-            self.__comp_state = None
+            self.__completer = None
+            self.__completion_seen = None
 
             # Start okuri-nasi conversion.
-            if keyval == u' ':
+            auto_start_henkan_keyword = None
+            rom_kana_state = tuple(self.__rom_kana_state)
+            rom_kana_state = self.__convert_rom_kana(keyval, rom_kana_state)
+            for keyword in AUTO_START_HENKAN_KEYWORDS:
+                if rom_kana_state[0].endswith(keyword):
+                    self.__auto_start_henkan_keyword = keyword
+                    break
+            if keyval == u' ' or self.__auto_start_henkan_keyword:
                 self.__conv_state = CONV_STATE_SELECT
                 self.__rom_kana_state = self.__convert_nn(self.__rom_kana_state)
                 midasi = self.__katakana_to_hiragana(self.__rom_kana_state[0])
-                self.__kana_kan_state = (midasi, None)
+                self.__midasi = midasi
                 usr_candidates = self.__usrdict.lookup(midasi)
                 sys_candidates = self.__sysdict.lookup(midasi)
                 candidates = self.__merge_candidates(usr_candidates,
                                                      sys_candidates)
                 self.__candidate_selector.set_candidates(candidates)
                 self.next_candidate()
-                return u''
+                return (True, u'')
 
             if is_shift and \
                     len(self.__rom_kana_state[1]) == 0 and \
                     not self.__okuri_rom_kana_state:
-                self.__okuri_rom_kana_state = \
-                    (u'', u'', self.__rom_kana_rule_tree)
+                self.__okuri_rom_kana_state = (u'', u'',
+                                               self.__rom_kana_rule_tree)
 
             if self.__okuri_rom_kana_state:
                 okuri = (self.__okuri_rom_kana_state[1] or keyval)[0]
@@ -853,14 +979,18 @@ class Context:
                     self.__conv_state = CONV_STATE_SELECT
                     midasi = self.__katakana_to_hiragana(\
                         self.__rom_kana_state[0] + okuri)
-                    self.__kana_kan_state = (midasi, None)
+                    self.__midasi = midasi
                     usr_candidates = self.__usrdict.lookup(midasi)
                     sys_candidates = self.__sysdict.lookup(midasi, okuri=True)
                     candidates = self.__merge_candidates(usr_candidates,
                                                          sys_candidates)
                     self.__candidate_selector.set_candidates(candidates)
                     self.next_candidate()
-                return u''
+                return (True, u'')
+
+            # Ignore ctrl+key and non-ASCII characters.
+            if is_ctrl or 0x20 > ord(letter) or ord(letter) > 0x7E:
+                return (False, u'')
 
             if self.__abbrev:
                 self.__rom_kana_state = (self.__rom_kana_state[0] + keyval,
@@ -868,23 +998,24 @@ class Context:
             else:
                 self.__rom_kana_state = \
                     self.__convert_rom_kana(keyval, self.__rom_kana_state)
-            return u''
+            return (True, u'')
 
         elif self.__conv_state == CONV_STATE_SELECT:
             if letter.isspace():
-                self.next_candidate()
+                if self.next_candidate() is None:
+                    self.__conv_state = CONV_STATE_START
+                return (True, u'')
             elif key == 'x':
-                self.previous_candidate()
+                if self.previous_candidate() is None:
+                    self.__conv_state = CONV_STATE_START
+                return (True, u'')
             elif key == 'ctrl+j' or key == 'return':
-                return self.kakutei()
-            else:
-                return self.kakutei() + self.press_key(key)
-            return u''
+                return (True, self.kakutei())
+            return (True, self.kakutei() + self.press_key(key)[1])
 
-    def __init_completion(self, compkey):
-        return (compkey,
-                itertools.chain(self.__comphist.get(compkey, list()),
-                                self.__sysdict.completer(compkey)))
+    def __init_completer(self, compkey):
+        return itertools.chain(self.__usrdict.completer(compkey),
+                               self.__sysdict.completer(compkey))
 
     def __delete_char_from_rom_kana_state(self, state):
         tree = self.__rom_kana_rule_tree
@@ -916,66 +1047,56 @@ class Context:
                 self.__rom_kana_state = state
                 return (True, u'')
         if self.__conv_state == CONV_STATE_START:
-            self.__conv_state = CONV_STATE_NONE
             self.reset()
-            self.activate_input_mode(INPUT_MODE_HIRAGANA)
             return (True, u'')
         return (False, u'')
 
-    def set_candidate_selector(self, candidate_selector):
-        '''Set candidate selector.'''
-        self.__candidate_selector = candidate_selector
-
-    def clear_candidates(self):
-        '''Clear the current candidates.'''
-        self.__candidate_selector.set_candidates(list())
-        self.__kana_kan_state = None
-        
     def next_candidate(self):
         '''Select the next candidate.'''
-        candidate = self.__candidate_selector.next_candidate()
-        self.__kana_kan_state = (self.__kana_kan_state[0], candidate)
+        return self.__candidate_selector.next_candidate()
             
     def previous_candidate(self):
         '''Select the previous candidate.'''
-        candidate = self.__candidate_selector.previous_candidate()
-        self.__kana_kan_state = (self.__kana_kan_state[0], candidate)
+        return self.__candidate_selector.previous_candidate()
 
     def __merge_candidates(self, usr_candidates, sys_candidates):
         return usr_candidates + \
             [candidate for candidate in sys_candidates
              if candidate not in usr_candidates]
 
-    def __preedit(self):
+    def split_preedit(self):
         if self.__conv_state == CONV_STATE_NONE:
             if self.__rom_kana_state:
-                return self.__rom_kana_state[1]
-            return u''
+                return (u'', self.__rom_kana_state[1], u'')
+            return (u'', u'', u'')
         elif self.__conv_state == CONV_STATE_START:
             if self.__okuri_rom_kana_state:
-                return u'▽' + self.__rom_kana_state[0] + u'*' + \
-                    self.__okuri_rom_kana_state[0] + \
-                    self.__okuri_rom_kana_state[1]
+                return (u'▽', self.__rom_kana_state[0],
+                        u'*' + self.__okuri_rom_kana_state[0] + \
+                            self.__okuri_rom_kana_state[1])
             else:
-                return u'▽' + self.__rom_kana_state[0] + \
-                    self.__rom_kana_state[1]
+                return (u'▽', self.__rom_kana_state[0] + \
+                            self.__rom_kana_state[1], u'')
         else:
             if self.__okuri_rom_kana_state:
-                if self.__kana_kan_state:
-                    candidate = self.__kana_kan_state[1]
+                if self.__midasi:
+                    candidate = self.__candidate_selector.candidate()
                     if candidate:
-                        return u'▼' + candidate[0] + \
-                            self.__okuri_rom_kana_state[0]
-                return u'▼' + self.__rom_kana_state[0] + \
-                    self.__okuri_rom_kana_state[0]
+                        return (u'▼', candidate[0],
+                                self.__okuri_rom_kana_state[0])
+                return (u'▼', self.__rom_kana_state[0],
+                        self.__okuri_rom_kana_state[0])
             else:
-                if self.__kana_kan_state:
-                    candidate = self.__kana_kan_state[1]
+                if self.__midasi:
+                    candidate = self.__candidate_selector.candidate()
                     if candidate:
-                        return u'▼' + candidate[0]
-                return u'▼' + self.__rom_kana_state[0]
+                        return (u'▼', candidate[0],
+                                (self.__auto_start_henkan_keyword or u''))
+                return (u'▼', self.__rom_kana_state[0],
+                        (self.__auto_start_henkan_keyword or u''))
+        return (u'', u'', u'')
 
-    preedit = property(lambda self: self.__preedit())
+    preedit = property(lambda self: ''.join(self.split_preedit()))
 
     def __convert_nn(self, state):
         output, pending, tree = state
@@ -984,7 +1105,7 @@ class Context:
                 output += u'ん'
             elif self.__input_mode == INPUT_MODE_KATAKANA:
                 output += u'ン'
-            return (output, pending, tree)
+            return (output, pending[:-1], tree)
         return state
         
     def __convert_rom_kana(self, letter, state):
